/*
**             Copyright 2017 by Kvaser AB, Molndal, Sweden
**                         http://www.kvaser.com
**
** This software is dual licensed under the following two licenses:
** BSD-new and GPLv2. You may use either one. See the included
** COPYING file for details.
**
** License: BSD-new
** ==============================================================================
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the <organization> nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
** AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
** LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
** CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
** SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
** BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
** IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
** ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
** POSSIBILITY OF SUCH DAMAGE.
**
**
** License: GPLv2
** ==============================================================================
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
**
**
** IMPORTANT NOTICE:
** ==============================================================================
** This source code is made available for free, as an open license, by Kvaser AB,
** for use with its applications. Kvaser AB does not accept any liability
** whatsoever for any third party patent or other immaterial property rights
** violations that may result from any usage of this source code, regardless of
** the combination of source code and various applications that it can be used
** in, or with.
**
** -----------------------------------------------------------------------------
*/

#include "common_defs.h"
#include "KvaLogWriter_Signal.h"
#include "TimeConv.h"
#include "SignalRow.h"

//===========================================================================
KvaLogWriter_Signal::KvaLogWriter_Signal()
{
  PRINTF(("KvaLogWriter_Signal::KvaLogWriter_Signal()\n"));
  sr = NULL; // created in write_header
  counter = 1;
}

//===========================================================================
KvaLogWriter_Signal::~KvaLogWriter_Signal()
{
  PRINTF(("KvaLogWriter_Signal::~KvaLogWriter_Signal()\n"));
  delete sr;
}

//===========================================================================
KvlcStatus KvaLogWriter_Signal::write_header()
{
  unsigned int i;
  unsigned int curColumn = 0;

  // Should only be done once
  if (!sr) {
    sr = new SignalRow(get_property_separator_char());
    for (i = 0; i < databases.size(); i++) {
      KvaDbHnd dh = databases[i];
      KvaDbMessageHnd mh;
      KvaDbSignalHnd sh;
      char msgname[500], signame[500], unit[500], fullname[500];

      KvaDbStatus dbstat = kvaDbGetFirstMsg(dh, &mh);
      if (dbstat != kvlcOK) {
        continue;
      }
      while (dbstat == kvlcOK) {
        KvaDbSignalHnd mux_signal;
        dbstat = kvaDbGetMsgName(mh, msgname, sizeof(msgname));
        if (dbstat != kvlcOK) {
          break;
        }
        dbstat = kvaDbGetMsgMux(mh, &mux_signal);
        if (dbstat != kvlcOK) {
          break;
        }
        dbstat = kvaDbGetFirstSignal(mh, &sh);
        if (dbstat != kvlcOK) {
          break;
        }
        while (dbstat == kvlcOK) {
          int mux_value = 0;
          dbstat = kvaDbGetSignalMode(sh, &mux_value);
          if (dbstat != kvlcOK) {
            PRINTF(("kvaDbGetSignalMode error %d\n", dbstat));
          }
          dbstat = kvaDbGetSignalName(sh, signame, sizeof(signame));
          if (dbstat != kvlcOK) {
            break;
          }
          dbstat = kvaDbGetSignalQualifiedName(sh, fullname, sizeof(fullname));
          if (dbstat != kvlcOK) {
            break;
          }

          sr->setName(curColumn, signame);
          sr->setFullName(curColumn, fullname);

          sr->setNumDecimals(curColumn, get_property_data_decimals());
          sr->setMux(curColumn, mux_value, mux_signal, reinterpret_cast<MuxCallback> (&MuxCallbackValueChecker));

          dbstat = kvaDbGetSignalUnit(sh, unit, sizeof(unit));
          if (dbstat != kvlcOK) {
            break;
          }

          sr->setUnit(curColumn, unit);

          double factor, offset;
          dbstat = kvaDbGetSignalValueScaling(sh, &factor, &offset);
          if (dbstat != kvlcOK) {
            break;
          }
          sr->setScaling(curColumn, factor, offset);

          sr->setMessageId(curColumn, mh);

          curColumn++;

          dbstat = kvaDbGetNextSignal(mh, &sh);
          if (dbstat != kvlcOK) {
            break;
          }
        }

        dbstat = kvaDbGetNextMsg(dh, &mh);
        if (dbstat != kvlcOK) {
          break;
        }
      }
    }

    if (get_property_show_counter()) {
      sr->setName(curColumn, "Counter");
      sr->setFullName(curColumn, "Counter");
      sr->setUnit(curColumn, "");
    }

    curColumn++;
  }
  return write_signal_header();
}

//===========================================================================
KvlcStatus KvaLogWriter_Signal::write_event(imLogData *logEvent)
{
  return write_row(logEvent);
}

//----------------------------------------------------------------------------------------
// forward any incoming imLogData as a sequence of
//   push(column, signalvalue, numdecimal)
// ending with a
//   commit(time)
// which states that all pushes since last commit happend at given time

// This means that it is possible to skip the commit until we find a message
// that is treated as a time reference, which gives us the Resampling function
// Werner wants.
KvlcStatus KvaLogWriter_Signal::write_row(imLogData *logEvent)
{
  unsigned int i;
  bool match = false;
  int curColumn = 0;
  KvlcStatus kvstatus = kvlcOK;


  if (!isOpened) {
    PRINTF(("KvaLogWriter_Signal::write_row, file is not opened\n"));
    return kvlcERR_FILE_ERROR;
  }

  if (!start_of_logging) setStartOfLogging(logEvent);

  switch (logEvent->common.type) {
    case ILOG_TYPE_MESSAGE: {
      time_int64 msgTime;

      if (!((1<<logEvent->msg.channel) & get_property_channel_mask())) {
        // dont use this channel so bail out
        return kvlcOK;
      }

      if (logEvent->msg.flags & canMSGERR_OVERRUN) {
        overrun_occurred = true;
      }

      if (logEvent->msg.flags & canMSG_ERROR_FRAME) {
        //length = sprintf(outbuffer, "%sErrorframe\n", outbuffer); match = true;
      }
      else {
        KvaDbHnd dh;
        KvaDbMessageHnd mh;
        KvaDbSignalHnd sh;
        unsigned int chanmask;

        msgTime = ((time_int64) logEvent->common.time64 +
                           (time_int64) get_property_offset());

        // time
        sr->setTimeStamp(msgTime); // High resolution timer
        sr->setAbsTime(logEvent->common.nanos_since_1970); // ABS time

        // make sure the id has the right format if we compare it to a j1939 db
        //PRINTF(("Got id: 0x%x -> ", logEvent->msg.id));
        if ((logEvent->msg.flags & canMSG_EXT) != 0) {
          // should this be here for non-j1939 as well? 
          logEvent->msg.id |= 0x80000000;
        }

        if (get_property_hlp_j1939()) {
          logEvent->msg.id = get_J1939_id(logEvent->msg.id);
        }

        for (i = 0; i < databases.size(); i++) {
          //
          // For each database dh
          //
          dh = databases[i];

          // get the channel-mask for this database....
          chanmask = databases_chanmask[i];

          KvaDbStatus dbstat = kvaDbGetFirstMsg(dh, &mh);
          if (dbstat != kvlcOK) {
            PRINTF(("kvadbgetfirstmsg failed %d\n", dbstat));
          }
          while (dbstat == kvlcOK) {
            //
            // for each message mh
            //
            unsigned int dbid;
            int          dbdlc;
            unsigned int dbflags;

            dbstat = kvaDbGetMsgId(mh, &dbid, &dbflags);
            if (dbstat != kvlcOK) {
              PRINTF(("kvaDbGetMsgId fel %d\n", dbstat));
            }

            dbstat = kvaDbGetMsgDlc(mh, &dbdlc);
            if (dbstat != kvlcOK) {
              PRINTF(("kvaDbGetMsgDlc fel %d\n", dbstat));
            }

            dbstat = kvaDbGetFirstSignal(mh, &sh);
            if (dbstat != kvlcOK) {
              PRINTF(("kvadbgetfirstsignal failed %d\n", dbstat));
            }
            while (dbstat == kvlcOK) {
              //
              // for each signal sh
              //
              if (dbid == logEvent->msg.id && (chanmask & (1 << (logEvent->msg.channel)))
                    && (sr->inEvent(curColumn, reinterpret_cast<unsigned char*>(logEvent->msg.data), logEvent->msg.dlc))) {
                //
                // correct ID
                //
                bool gotSignalValue = false;
                if (get_property_enum_values()) {
                  char buf[MAX_ELEM_LEN];
                  dbstat = kvaDbGetSignalValueEnum(sh, buf, sizeof(buf),
                                                  logEvent->msg.data,
                                                  logEvent->msg.dlc);
                  if (dbstat == kvlcOK) {
                    gotSignalValue = true;
                    sr->setSignalValue(curColumn, buf);
                  }
                }
                if (!gotSignalValue) {
                  double  valdbl;

                  dbstat = kvaDbGetSignalValueFloat(sh, &valdbl,
                  logEvent->msg.data,
                  logEvent->msg.dlc);
                  if (dbstat != kvlcOK) {
                    PRINTF(("write_row: kvaDbGetSignalValueFloat failed\n"));
                  }

                  sr->setSignalValue(curColumn, valdbl);
                }
                match = true;
              } // end correct id
              curColumn++;
              dbstat = kvaDbGetNextSignal(mh, &sh);
              if (dbstat != kvlcOK) {
                break;
              }
            } // end for each signal

            dbstat = kvaDbGetNextMsg(dh, &mh);
            if (dbstat != kvlcOK) {
              break;
            }
          } // end for each message
        } // end for each database

        if (get_property_show_counter()) {
          sr->setSignalValue(curColumn, (double)counter);
          sr->setNumDecimals(curColumn, 0);
        }
        curColumn++;

      }

      if (match) {
        bool write = true;
        // int col;


        if (get_property_merge_lines()) {
          write = sr->isUpdated();
        }

        /*if (-1 != (col = get_property_resample_column())) {
          write = sr->isAccessed(col);
        }*/

        if (write) {
          kvstatus = write_signals();
          if (kvstatus != kvlcOK) return kvstatus;

          if ( split_on_time(msgTime) ) {
            kvstatus = split_file();
          }
          // Since we have printed the row, we reset its updated-status
          sr->resetUpdated();
        }
        counter++;

      }
      break;
    }

    case ILOG_TYPE_TRIGGER:
    case ILOG_TYPE_RTC:
    case ILOG_TYPE_VERSION:
    {
      break;
    }

    default:
    {
      PRINTF(("logEvent->common.type = %d, frame_counter = %d\n",
              logEvent->common.type,
              logEvent->common.event_counter));
      break;
    }

  }
  logEvent->common.new_data = false;


  return kvstatus;
}

bool KvaLogWriter_Signal::get_property_full_names()
{
  return (mProperties.mShowFullyQualifiedNames == PROPERTY_ON);
}
